package config

import (
	"github.com/stretchr/testify/assert"
	"testing"
)

func InitFileTest() {
	KeenTune = new(KeentunedConf)
	KeenTune.Home = "/etc/keentune/keentuned"
	KeenTune.DumpHome = "/var/keentune/keentuned"
}
func Test_checkBenchConf(t *testing.T) {
	InitFileTest()
	var (
		fileConf = "wrk_http_long.json"
	)

	ret := checkBenchConf(&fileConf)
	assert.Nil(t, ret)
}

func Test_checkItem(t *testing.T) {
	var (
		value = map[string]interface{}{
			"items": map[string]interface{}{
				"Requests_sec": map[string]interface{}{
					"negative": false,
					"weight":   float64(100),
					"strict":   false,
				},
				"Transfer_sec": map[string]interface{}{
					"negative": false,
					"weight":   float64(0),
					"strict":   false,
				},
				"Latency_90": map[string]interface{}{
					"negative": true,
					"weight":   float64(0),
					"strict":   true,
				},
				"Latency_99": map[string]interface{}{
					"negative": true,
					"weight":   float64(0),
					"strict":   true,
				},
			},
		}
	)

	err := checkItem(value)
	assert.Nil(t, err)
}

func Test_parse2String(t *testing.T) {
	var (
		value = map[string]interface{}{
			"aaa": "hello",
			"bbb": "world",
		}
		key = "aaa"
	)

	ret := parse2String(value, key)
	assert.Nil(t, ret)
}

func Test_parse2Bool(t *testing.T) {
	var (
		value = map[string]interface{}{
			"aaa": true,
			"bbb": "world",
		}
		key = "aaa"
	)

	ret := parse2Bool(value, key)
	assert.Nil(t, ret)
}

func Test_parse2Float(t *testing.T) {
	var (
		value = map[string]interface{}{
			"aaa": float64(32),
			"bbb": "world",
		}
		key = "aaa"
	)

	_, ret := parse2Float(value, key)
	assert.Nil(t, ret)
}

func Test_checkParamConf(t *testing.T) {
	InitFileTest()
	var (
		fileConf = "sysctl.json"
		gpNo     = 1
		gp       = &Group{}
	)

	ret := checkParamConf([]string{fileConf}, gpNo, gp)
	assert.Nil(t, ret)
}

func Test_readFile(t *testing.T) {
	var (
		fileName = "/etc/keentune/keentuned/parameter/sysctl.json"
		gpNo     = 1
	)

	_, _, ret := readFile(fileName, gpNo)
	assert.Nil(t, ret)
}

func Test_parseRules(t *testing.T) {
	var (
		param = []interface{}{
			[]interface{}{
				"client_body_timeout@group-1",
				"send_timeout",
				">=",
			},
		}

		rules = make([][3]string, 0)
	)

	ret := parseRules(param, &rules)
	assert.Nil(t, ret)
}

func Test_readParams(t *testing.T) {
	var (
		param = DBLMap{
			"sysctl": {
				"vm.page-cluster": map[string]interface{}{
					"domain": "sysctl",
					"range": []interface{}{
						0, 8,
					},
					"dtype": "int",
					"step":  1,
					"base":  "6",
					"value": 5,
				},
			},
		}
		domain = map[string]string{}
		merged = make([]DBLMap, PRILevel)
	)

	ret := readParams(domain, param, merged)
	assert.Nil(t, ret)
}

func Test_checkParam(t *testing.T) {
	var (
		name  = "vm.page-cluster"
		param = map[string]interface{}{
			"domain": "sysctl",
			"range": []interface{}{
				0, 8,
			},
			"dtype": "int",
			"step":  1,
			"base":  "6",
			"value": 5,
		}
	)

	ret := checkParam(name, param)
	assert.Nil(t, ret)
}

func Test_checkUniqueField(t *testing.T) {
	var (
		paramNil = Parameter{
			Dtype: "string",
		}
		paramBothScopeAndSequence = Parameter{
			Dtype: "string",
			Scope: []interface{}{
				1, 8,
			},
			Sequence: []interface{}{
				2, 6,
			},
		}
		paramBothScopeAndOptions = Parameter{
			Dtype: "string",
			Scope: []interface{}{
				1, 8,
			},
			Options: []string{
				"0", "1",
			},
		}
		paramBothOptionsAndSequence = Parameter{
			Dtype: "string",
			Sequence: []interface{}{
				2, 6,
			},
			Options: []string{
				"0", "1",
			},
		}
		paramDtypeErr = Parameter{
			Dtype: "string",
			Step:  1,
		}
	)

	ret1 := checkUniqueField(paramNil)
	assert.NotNil(t, ret1)

	ret2 := checkUniqueField(paramBothScopeAndSequence)
	assert.NotNil(t, ret2)

	ret3 := checkUniqueField(paramBothScopeAndOptions)
	assert.NotNil(t, ret3)

	ret4 := checkUniqueField(paramBothOptionsAndSequence)
	assert.Nil(t, ret4)

	ret5 := checkUniqueField(paramDtypeErr)
	assert.NotNil(t, ret5)
}

func Test_isDataTypeOK(t *testing.T) {
	var (
		strType   = "string"
		boolType  = "bool"
		intType   = "int"
		floatType = "float"
		other     = ""

		isStr   = true
		isBool  = true
		isInt   = true
		isFloat = true
		notOK   = false
	)

	ret1 := isDataTypeOK(strType)
	assert.Equal(t, isStr, ret1)

	ret2 := isDataTypeOK(boolType)
	assert.Equal(t, isBool, ret2)

	ret3 := isDataTypeOK(intType)
	assert.Equal(t, isInt, ret3)

	ret4 := isDataTypeOK(floatType)
	assert.Equal(t, isFloat, ret4)

	ret5 := isDataTypeOK(other)
	assert.Equal(t, notOK, ret5)
}

func TestGetPriorityParams(t *testing.T) {
	var (
		userParamMap = DBLMap{
			"sysctl": {
				"vm.page-cluster": map[string]interface{}{
					"domain": "sysctl",
					"range": []interface{}{
						0, 8,
					},
					"dtype": "int",
					"step":  1,
					"base":  "6",
					"value": 5,
				},
			},
		}
	)

	_, ret := GetPriorityParams(userParamMap)
	assert.Nil(t, ret)
}

func TestGetKeenTuneConfFileMD5(t *testing.T) {
	ret := GetKeenTuneConfFileMD5()
	assert.NotEmpty(t, ret)
}

