/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package config for daemon, this package contains the check, config, priority, workpath. Parse the keentuned.conf file to get the ip, port information of the other three components.Determine whether the profile information meets and provide the file path related to keentune.
 */

package config

import (
	"encoding/json"
	"testing"

	"github.com/go-ini/ini"
	"github.com/stretchr/testify/assert"
)

func InitNodesTest() {
	KeenTune = new(KeentunedConf)
	KeenTune.LogFileLvl = "INFO"
	KeenTune.FileName = "keentuned.log"
	KeenTune.Interval = 7
	KeenTune.BackupCount = 14
	KeenTune.Home = "/etc/keentune/keentuned"
	KeenTune.DumpHome = "/var/keentune/keentuned"

	InitBenchTest()
	InitTargetTest()
	RealLocalIP = "localhost"
	KeenTune.BrainPort = "9872"
	KeenTune.BrainIP = "localhost"
	KeenTune.HeartbeatTime = 3
	RealLocalIP = "localhost"
	KeenTune.Port = "9871"

	KeenTune.BaseRound = 1
	KeenTune.ExecRound = 1
	KeenTune.AfterRound = 1

	KeenTune.SensTimeout = 1
	KeenTune.TargetSetTimeout = 1
	KeenTune.BenchTimeout = 1
}

func InitBenchTest() {
	var benchGp []BenchGroup
	benchGp = append(benchGp, BenchGroup{
		SrcPort: "9874",
		SrcIPs:  []string{"localhost"},
		DestIP:  "localhost",
	})

	bencIpMap := map[string]int{
		"localhost": 1,
	}
	KeenTune.Bench = Bench{
		BenchGroup: benchGp,
		BenchIPMap: bencIpMap,
	}
}

func InitTargetTest() {
	var targetGp Target
	var paramStr = `{
	"sysctl": {
		"vm.page-cluster": {
			"dtype": "int"
		},
		"vm.stat_interval": {
			"dtype": "int"
		},
		"vm.swappiness": {
			"dtype": "int"
		},
		"vm.vfs_cache_pressure": {
			"dtype": "int"
		},
		"vm.watermark_scale_factor": {
			"dtype": "int"
		},
		"vm.zone_reclaim_mode": {
			"dtype": "string"
		}
	}
}`
	var (
		paramMaps []DBLMap
		params    DBLMap
	)

	json.Unmarshal([]byte(paramStr), &params)
	paramMaps = append(paramMaps, params)

	targetGp = Target{
		Group: []Group{
			{
				IPs:      []string{"localhost"},
				Port:     "9873",
				GroupNo:  1,
				ParamMap: paramMaps,
			},
		},
		IPMap: map[string]int{
			"localhost": 1,
		},
	}

	KeenTune.Target = targetGp
}
func TestInit(t *testing.T) {
	Init()
	assert.Nil(t, nil)
}

func Test_initChanAndIPMap(t *testing.T) {
	InitNodesTest()
	initChanAndIPMap()
	assert.Nil(t, nil)
}

func TestKeentunedConf_Save(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
	)

	ret := testKeenTune.Save()
	assert.Nil(t, ret)
}

func TestKeentunedConf_getDefault(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
	)

	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	testKeenTune.getDefault(cfg)
	assert.Nil(t, nil)
}

func TestKeentunedConf_getTargetGroup(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
	)

	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	ret := testKeenTune.getTargetGroup(cfg)
	assert.Nil(t, ret)
}

func Test_hasGroupSections(t *testing.T) {
	var groupNames = make([]string, 0)
	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	ret := hasGroupSections(cfg, &groupNames, BenchSectionPrefix)
	assert.Equal(t, true, ret)
}

func TestKeentunedConf_getBenchGroup(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
	)

	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	ret := testKeenTune.getBenchGroup(cfg)
	assert.Nil(t, ret)
}

func Test_checkIPRepeated(t *testing.T) {
	var (
		ips = []string{
			"localhost",
			"127.0.0.1",
			"100.81.214.61",
			"39.104.161.37",
			"100.81.214.61",
		}
		gpName      = "target-group-1"
		allGroupIPs = map[string]string{}
	)

	err := checkIPRepeated(gpName, ips, allGroupIPs)
	assert.NotNil(t, err)
}

func TestKeentunedConf_GetLogConf(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
	)

	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	testKeenTune.GetLogConf(cfg)
	assert.Nil(t, nil)
}

func TestKeentunedConf_addTargetIPMap(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
		ips          = []string{
			"127.0.0.1",
			"100.81.214.61",
			"39.104.161.37",
		}
		ipExist = make(map[string]bool)
		id      = new(int)
	)

	testKeenTune.Target.IPMap = make(map[string]int)

	testKeenTune.addTargetIPMap(ips, ipExist, id)
	assert.Nil(t, nil)
}

func TestKeentunedConf_addBenchIPMap(t *testing.T) {
	var (
		testKeenTune = &KeentunedConf{}
		ips          = []string{
			"127.0.0.1",
			"100.81.214.61",
		}
		ipExist = make(map[string]bool)
		id      = new(int)
	)

	testKeenTune.Bench.BenchIPMap = make(map[string]int)

	testKeenTune.addBenchIPMap(ips, ipExist, id)
	assert.Nil(t, nil)
}

func Test_changeStringToSlice(t *testing.T) {
	var (
		ipStr = "39.104.161.37, 100.81.214.61"
	)

	ret, _ := changeStringToSlice(ipStr)
	assert.NotNil(t, ret)
}

func TestInitWorkDir(t *testing.T) {
	ret := InitWorkDir()
	assert.Nil(t, ret)
}

func Test_getWorkDir(t *testing.T) {
	cfg, _ := ini.InsensitiveLoad(keentuneConfigFile)
	getWorkDir(cfg)
	assert.Nil(t, nil)
}

func TestInitTargetGroup(t *testing.T) {
	InitTargetGroup()
	assert.Nil(t, nil)
}

func TestGetJobParamConfig(t *testing.T) {
	InitFileTest()
	var job = "05_04"
	_, _, err := GetJobParamConfig(job)
	assert.Nil(t, err)
}

func TestGetKeenTunedYaml(t *testing.T) {
	ret := GetKeenTunedYaml()
	assert.NotEmpty(t, ret)
}
