/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package file for daemon, this package contains the file. The package implements the processing of files, including the transformation of file types, path acquisition, and directory existence judgment.
 */

package file

import (
	"github.com/stretchr/testify/assert"
	"os"
	"testing"
)

func TestIsPathExist(t *testing.T) {
	var (
		existFile = "/etc/keentune/keentuned"
		noneExist = "aaa.conf"
	)

	got := IsPathExist(existFile)
	assert.Equal(t, true, got)

	got2 := IsPathExist(noneExist)
	assert.Equal(t, false, got2)
}

func TestDecoratePath(t *testing.T) {
	var (
		empty         = ""
		withSuffix    = "aaa/"
		withoutSuffix = "aaa"
	)

	ret := DecoratePath(empty)
	assert.Empty(t, ret)

	ret1 := DecoratePath(withSuffix)
	assert.NotEmpty(t, ret1)

	ret2 := DecoratePath(withoutSuffix)
	assert.NotEmpty(t, ret2)
}

func TestReadFile2Map(t *testing.T) {
	var (
		fileName = "/etc/keentune/keentuned/parameter/sysctl.json"
	)

	_, err := ReadFile2Map(fileName)
	assert.Nil(t, err)
}

func TestDump2File(t *testing.T) {
	var (
		path     = "."
		fileName = "aaa.txt"
		info     = map[string]interface{}{
			"aaa": "hello world",
		}
	)
	err := Dump2File(path, fileName, info)
	assert.Nil(t, err)
	os.Remove(fileName)
}

func TestWalkFilePath(t *testing.T) {
	var (
		path  = "."
		match = ".go"
	)

	_, _, ret := WalkFilePath(path, match)
	assert.Nil(t, ret)
}

func TestSave2CSV(t *testing.T) {
	var (
		path = "."
		name = "bbb.csv"
		data = map[string][]float32{
			"a": {
				float32(1), float32(2),
			},
			"x": {
				float32(2.3), float32(4.1),
			},
		}
	)

	err := Save2CSV(path, name, data)
	assert.Nil(t, err)
	os.Remove(name)
}

func TestGetWalkPath(t *testing.T) {
	var (
		name = "file_test.go"
		path = "./"
	)

	walkPath, _ := GetWalkPath(path, name)
	assert.NotEmpty(t, walkPath)
}

func TestGetPlainName(t *testing.T) {
	var (
		name     = "/etc/keentune/keentuned.conf"
		expected = "keentuned.conf"
	)

	plainName := GetPlainName(name)
	assert.Equal(t, expected, plainName)
}

func TestCopy(t *testing.T) {
	var (
		src  = "./file_test.go"
		dest = "./file_test.txt"
	)

	err := Copy(src, dest)
	assert.Nil(t, err)
	os.Remove(dest)
}
