/* 
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd. 
* SPDX-License-Identifier: MulanPSL-2.0 
*/ 
package file

import (
	"encoding/csv"
	"fmt"
	"github.com/pkg/errors"
	"os"
	"sync"
	"syscall"
)

var (
	Empty         = errors.New("record is empty")
	NotExist      = errors.New("record not found")
	HeaderIsEmpty = errors.New("header is empty")
	AlreadyExist  = errors.New("job or record is already existence")
)

// CreatCSV create csv file with header
func CreatCSV(fileName string, header []string) error {
	if len(header) == 0 {
		return HeaderIsEmpty
	}

	var mutex = &sync.RWMutex{}
	mutex.Lock()
	defer mutex.Unlock()

	return addRecord(fileName, header)

}

func addRecord(fileName string, header []string) error {
	f, err := os.OpenFile(fileName, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0666)
	if err != nil {
		return err
	}

	defer f.Close()
	w := csv.NewWriter(f)
	err = w.Write(header)
	if err != nil {
		return err
	}

	w.Flush()
	return nil
}

// Append append for csv file
func Append(fileName string, record []string) error {
	f, err := os.OpenFile(fileName, os.O_WRONLY|os.O_APPEND, 0666)
	if err != nil {
		return err
	}

	defer f.Close()
	w := csv.NewWriter(f)
	err = w.Write(record)
	if err != nil {
		return err
	}

	w.Flush()
	return nil
}

// Insert Null value must be assigned "-"
func Insert(fileName string, record []string, prKey ...int) error {
	if len(record) == 0 {
		return Empty
	}

	var mutex = &sync.RWMutex{}
	mutex.Lock()
	defer mutex.Unlock()

	if len(prKey) > 0 {
		if HasRecord(fileName, record[prKey[0]], prKey[0]) {
			return fmt.Errorf("'%v' %v", record[0], AlreadyExist)
		}
	}

	return Append(fileName, record)
}

// UpdateRow update row info
func UpdateRow(fileName string, prKey int, info map[int]string) error {
	var mutex = &sync.RWMutex{}
	mutex.RLock()
	defer mutex.RUnlock()

	records, err := GetAllRecords(fileName)
	if err != nil {
		return err
	}

	prVal := info[prKey]

	var found bool
	for rowIdx, record := range records {
		if record[prKey] == prVal {
			found = true
			for colIdx, value := range info {
				records[rowIdx][colIdx] = value
			}
		}
	}

	if !found {
		return NotExist
	}

	return truncWrite(fileName, records)
}

// DeleteRow delete row by primary key
func DeleteRow(fileName string, prVal string, prKey int) error {
	var mutex = &sync.RWMutex{}
	mutex.RLock()
	defer mutex.RUnlock()

	records, err := GetAllRecords(fileName)
	if err != nil || len(records) == 0 {
		return err
	}

	if prKey < 0 || prKey >= len(records[0]) {
		return fmt.Errorf("index %v out of range [0, %v)", prKey, len(records[0]))
	}

	var found bool

	var newContents [][]string
	for _, record := range records {
		if len(record) == 0 {
			continue
		}

		if record[prKey] == prVal {
			found = true
			continue
		}

		newContents = append(newContents, record)
	}

	if !found {
		return NotExist
	}

	return truncWrite(fileName, newContents)
}

// IsInSlice ...
func IsInSlice(obj string, pond []string) bool {
	for _, item := range pond {
		if obj == item {
			return true
		}
	}

	return false
}

// HasRecord ...
func HasRecord(fileName, match string, key int) bool {
	records, err := GetAllRecords(fileName)
	if len(records) == 0 || err != nil {
		return false
	}

	if key < 0 || key >= len(records[0]) {
		return false
	}

	for _, record := range records {
		if len(record) == 0 {
			continue
		}

		if record[key] == match {
			return true
		}
	}

	return false
}

// IsJobRunning assert job status equal running...
func IsJobRunning(fileName, val1 string, key1, key2 int) bool {
	status := GetRecord(fileName, val1, key1, key2)
	return status == "running"
}

// GetRecord ...
func GetRecord(fileName, value1 string, key1, key2 int) string {
	record, err := GetOneRecord(fileName, value1, key1)
	if err != nil || len(record) == 0 {
		return ""
	}

	if key2 < 0 || key2 >= len(record) {
		return ""
	}

	return record[key2]
}

// GetAllRecords ...
func GetAllRecords(fileName string) ([][]string, error) {
	file, err := os.Open(fileName)
	if err != nil {
		return nil, NotExist
	}

	defer file.Close()

	r := csv.NewReader(file)
	
	return r.ReadAll()
}

// GetOneRecord ...
func GetOneRecord(fileName, matched string, primaryKey int) ([]string, error) {
	records, err := GetAllRecords(fileName)
	if err != nil || len(records) == 0 {
		return nil, err
	}

	if primaryKey < 0 || primaryKey >= len(records[0]) {
		return nil, fmt.Errorf("index %v out of range [0, %v)", primaryKey, len(records[0]))
	}

	for i, record := range records {
		if len(record) == 0 {
			continue
		}
		if record[primaryKey] == matched {
			return records[i], nil
		}
	}

	return nil, NotExist
}

func truncWrite(fileName string, info [][]string) error {
	f, err := os.OpenFile(fileName, os.O_WRONLY|syscall.O_TRUNC, 0666)
	if err != nil {
		return err
	}

	defer f.Close()
	w := csv.NewWriter(f)
	err = w.WriteAll(info)
	if err != nil {
		return err
	}

	w.Flush()
	return nil
}

