/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package log for daemon, this package contains the log. The package defines the command name, defines the log level, mainly using warning, error, info three levels, to achieve the log output screen and file two ways.
 */

package log

import (
	"github.com/stretchr/testify/assert"
	"keentune/daemon/common/config"
	"keentune/daemon/common/utils"
	"os"
	"testing"
	"time"

	"github.com/sirupsen/logrus"
)

func initLogTest() {
	config.KeenTune = new(config.KeentunedConf)
	config.KeenTune.LogFileLvl = "DEBUG"

	config.KeenTune.Interval = 2
	config.KeenTune.BackupCount = 14
	config.KeenTune.FileName = "keentuned.log"
}
func Test_getLevel(t *testing.T) {
	var (
		lvlUnExpected = "aaa"
		lvlInfo       = "info"
	)

	got := getLevel(lvlUnExpected)
	assert.Equal(t, logrus.InfoLevel, got)

	got = getLevel(lvlInfo)
	assert.Equal(t, logrus.InfoLevel, got)
}

func TestInit(t *testing.T) {
	initLogTest()
	Init()
	assert.Nil(t, nil)
}

func Test_logFormater_Format(t *testing.T) {
	var (
		ins = logFormater{
			TimestampFormat: time.Now().Local().Format("2006-01-02 15:04:05"),
			LogFormat:       "PID= %v [%s] %s %s ...  [%s, %d]\n",
			file:            "daemon",
			funcName:        "log",
			line:            1,
			pid:             os.Getpid(),
		}

		logger = logrus.New()

		fields = logrus.Fields{
			"foo": "bar",
			"baz": "qux",
		}

		entry = &logrus.Entry{
			Time:    time.Time{},
			Level:   logrus.DebugLevel,
			Message: "message",
			Data:    fields,
			Logger:  logger,
		}
	)

	info, _ := ins.Format(entry)
	assert.NotEmpty(t, info)
}

func Test_removeColor(t *testing.T) {
	var (
		colorStr = utils.ColorString("Red", "hello")
		expected = "hello"
		pureStr  = "hello"
	)

	ret := removeColor(colorStr)
	assert.Equal(t, expected, ret)

	ret = removeColor(pureStr)
	assert.Equal(t, expected, ret)
}

func Test_consoleLogFormater_Format(t *testing.T) {
	var (
		ins = consoleLogFormater{
			LogFormat: "[%s] %s\n",
		}
		logger = logrus.New()

		fields = logrus.Fields{
			"foo": "bar",
			"baz": "qux",
		}

		entry = &logrus.Entry{
			Time:    time.Time{},
			Level:   logrus.DebugLevel,
			Message: "message",
			Data:    fields,
			Logger:  logger,
		}
	)

	bytes, _ := ins.Format(entry)
	assert.NotNil(t, bytes)
}

func Test_updateClientLog(t *testing.T) {
	var (
		cmd1 = ParamDel
		cmd2 = "param tune" + ":aaa.log"
		msg  = "step1. do something ..."
	)

	updateClientLog(cmd1, msg)
	assert.Nil(t, nil)

	updateClientLog(cmd2, msg)
	assert.Nil(t, nil)
	os.Remove("aaa.log")
	ClearCliLog(cmd1)
}

func TestClearCliLog(t *testing.T) {
	var (
		cmd = ParamDump
	)

	ClearCliLog(cmd)
	assert.Nil(t, nil)
}

func TestLogger_setFormatter(t *testing.T) {
	initLogTest()
	Init()
	var (
		modeFile    = "file"
		modeConsole = "console"
	)

	ret := fLogInst.setFormatter(modeFile)
	assert.NotNil(t, ret)

	ret = cLogInst.setFormatter(modeConsole, "warning")
	assert.NotNil(t, ret)
}

func Test_getFile(t *testing.T) {
	var (
		name1 = "daemon/common/common.go"
		name2 = "daemon.go"
	)

	ret1 := getFile(name1)
	assert.NotEmpty(t, ret1)

	ret2 := getFile(name2)
	assert.NotEmpty(t, ret2)
}

func TestLogger_getFormater(t *testing.T) {
	initLogTest()
	Init()
	var (
		file     = "daemon/common/log"
		funcName = "fetFormat"
		line     = 256
		lvlDebug = "DEBUG"
		lvlErr   = "ERROR"
	)

	ret := cLogInst.getFormater(file, funcName, line, lvlDebug)
	assert.NotEmpty(t, ret.LogFormat)

	ret = cLogInst.getFormater(file, funcName, line, lvlErr)
	assert.NotEmpty(t, ret.LogFormat)
}

func TestInfo(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Info log test\n"
	)

	Info(cmd, msg)
	assert.Nil(t, nil)
}

func TestInfoln(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Infoln log test"
	)

	Infoln(cmd, msg)
	assert.Nil(t, nil)
}

func TestInfof(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd    = ""
		msg    = "this is Infof log test"
		format = "%v\n"
	)

	Infof(cmd, format, msg)
	assert.Nil(t, nil)
}

func TestWarn(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Warn log test\n"
	)

	Warn(cmd, msg)
	assert.Nil(t, nil)
}

func TestWarnln(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Warnln log test"
	)

	Warnln(cmd, msg)
	assert.Nil(t, nil)
}

func TestWarnf(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd    = ""
		msg    = "this is Warnf log test"
		format = "%v\n"
	)

	Warnf(cmd, format, msg)
	assert.Nil(t, nil)
}

func TestError(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Error log test\n"
	)

	Error(cmd, msg)
	assert.Nil(t, nil)
}

func TestErrorln(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Errorln log test"
	)

	Errorln(cmd, msg)
	assert.Nil(t, nil)
}

func TestErrorf(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd    = ""
		msg    = "this is Errorf log test"
		format = "%v\n"
	)

	Errorf(cmd, format, msg)
	assert.Nil(t, nil)
}

func TestDebug(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Debug log test\n"
	)

	Debug(cmd, msg)
	assert.Nil(t, nil)
}

func TestDebugln(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd = ""
		msg = "this is Debugln log test"
	)

	Debugln(cmd, msg)
	assert.Nil(t, nil)
}

func TestDebugf(t *testing.T) {
	initLogTest()
	Init()

	var (
		cmd    = ""
		msg    = "this is Debugf log test"
		format = "%v\n"
	)

	Debugf(cmd, format, msg)
	assert.Nil(t, nil)
}

func Test_cacheLog(t *testing.T) {
	var (
		cmd      = ProfDel
		lvlInfo  = "info"
		lvlError = "error"

		format string
		arg    = "this is cache log test"
	)

	cacheLog(cmd, lvlInfo, format, arg)
	assert.Nil(t, nil)

	cacheLog(cmd, lvlError, format, arg)
	assert.Nil(t, nil)

}

func Test_colorLevel(t *testing.T) {
	var (
		lvlE         = "ERROR"
		lvlW         = "WARNING"
		lvlInfo      = "info"
		expextedE    = "\u001B[1;40;31mERROR\u001B[0m"
		expextedW    = "\u001B[1;40;33mWARNING\u001B[0m"
		expextedInfo = lvlInfo
	)

	ret := colorLevel(lvlE)
	assert.Equal(t, expextedE, ret)

	ret = colorLevel(lvlW)
	assert.Equal(t, expextedW, ret)

	ret = colorLevel(lvlInfo)
	assert.Equal(t, expextedInfo, ret)
}
