/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package http for daemon, this package contains the http. The package contains the remote call the url with the specified method and request data, get success response after request url, and parses the return data.
 */
package http

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"keentune/daemon/common/config"
	"keentune/daemon/common/log"
	"net"
	"net/http"
	"strings"
	"time"
)

// Protocol  ...
const Protocol = "http"

// StatusOk ...
const StatusOk = 200

const initDomainURI = "9873/init"

type requester struct {
	client *http.Client
	url    string
	body   []byte
	method string
}

func newRequester(method string, uri string, data interface{}) (*requester, error) {
	url := Protocol + "://" + uri
	timeout := 5 * time.Minute
	if strings.Contains(uri, initDomainURI) {
		timeout = time.Duration(config.KeenTune.TargetSetTimeout) * time.Minute
	}

	var client = &http.Client{
		Transport: &http.Transport{
			DialContext: (&net.Dialer{
				Timeout: 900 * time.Millisecond, // conn timeout
			}).DialContext,
			DisableKeepAlives:     true,
			ResponseHeaderTimeout: timeout,
		},
	}

	if data == nil && method == "GET" {
		log.Debugf("", "Request [%v]", url)
		return &requester{
			client: client,
			url:    url,
			body:   nil,
			method: method,
		}, nil
	}

	// Avoid escaping special characters such as "<" and ">"
	bf := bytes.NewBuffer([]byte{})
	jsonEncoder := json.NewEncoder(bf)
	jsonEncoder.SetEscapeHTML(false)
	err := jsonEncoder.Encode(data)

	if err != nil {
		fmt.Printf("[newRequester] error in Unexpected type %T\n", data)
		return nil, err
	}

	log.Debugf("", "[%v] request info [%+v]", url, bf.String())
	return &requester{
		client: client,
		url:    url,
		body:   bf.Bytes(),
		method: method,
	}, nil
}

func (r *requester) execute() (*http.Response, error) {
	request, err := http.NewRequest(r.method, r.url, bytes.NewReader(r.body))

	if err != nil {
		return nil, err
	}

	if r.body != nil {
		request.Header.Set("Content-Type", "application/json")
	}

	response, err := r.client.Do(request)
	if err != nil {
		return nil, err
	}

	return response, nil
}

// RemoteCall remote call the url with the specified method and request data
func RemoteCall(method, url string, data interface{}) ([]byte, error) {
	requester, err := newRequester(method, url, data)
	if err != nil {
		return nil, err
	}

	response, err := requester.execute()
	if err != nil {
		return nil, err
	}

	if response.Body == nil {
		return nil, fmt.Errorf("[%v] response is nil", method+url)
	}

	defer response.Body.Close()

	bytesData, err := ioutil.ReadAll(response.Body)
	if err != nil {
		return nil, err
	}

	if response.StatusCode != StatusOk {
		return nil, fmt.Errorf("%s", parseMessage(bytesData))
	}

	log.Debugf("", "[%v] response info [%+v]", Protocol+"://"+url, string(bytesData))
	return bytesData, nil
}

// ResponseSuccess get success response after request url
func ResponseSuccess(method, url string, request interface{}) error {
	resp, err := RemoteCall(method, url, request)
	if err != nil {
		return err
	}

	message := parseMessage(resp)
	if message != "" {
		return fmt.Errorf("response suc is false, msg is %v", message)
	}

	return nil
}

func parseMessage(resp []byte) string {
	var response struct {
		Success bool        `json:"suc"`
		Message interface{} `json:"msg"`
	}

	if err := json.Unmarshal(resp, &response); err != nil {
		return string(resp)
	}

	if !response.Success {
		return fmt.Sprint(response.Message)
	}

	return ""
}
