/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package http for daemon, this package contains the http. The package contains the remote call the url with the specified method and request data, get success response after request url, and parses the return data.
 */

package http

import (
	"keentune/daemon/common/config"
	"keentune/daemon/common/log"
	"net"
	"net/http"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func initHttpTest() {
	config.KeenTune = new(config.KeentunedConf)
	config.KeenTune.TargetSetTimeout = 1
	config.KeenTune.LogFileLvl = "DEBUG"

	config.KeenTune.Interval = 2
	config.KeenTune.BackupCount = 14
	config.KeenTune.FileName = "keentuned.log"
	log.Init()
}
func Test_newRequester(t *testing.T) {
	initHttpTest()
	var (
		get       = "GET"
		post      = "POST"
		statusURI = "localhost:9873/avaliable"
		initURI   = "localhost:9873/init"
		data1     interface{}
		data2     = `
{
    "domain_list": ["sysctl"],
    "backup_all": false
}`
	)

	_, err := newRequester(get, statusURI, data1)
	assert.Nil(t, err)

	_, err = newRequester(post, initURI, data2)
	assert.Nil(t, err)
}

func Test_requester_execute(t *testing.T) {
	initHttpTest()

	var (
		timeout = 15 * time.Second
		client  = &http.Client{
			Transport: &http.Transport{
				DialContext: (&net.Dialer{
					Timeout: 900 * time.Millisecond,
				}).DialContext,
				DisableKeepAlives:     true,
				ResponseHeaderTimeout: timeout,
			},
		}

		get = requester{
			client: client,
			url:    Protocol + "://localhost:9873/avaliable",
			body:   nil,
			method: "GET",
		}
		post = requester{
			client: client,
			url:    Protocol + "://localhost:9873/init",
			body: []byte(`
					{
   					 	"domain_list": ["sysctl"],
						"backup_all": false
					}
            `),
			method: "POST",
		}
	)

	_, err := get.execute()
	assert.Nil(t, err)

	_, err = post.execute()
	assert.Nil(t, err)
}

func TestRemoteCall(t *testing.T) {
	initHttpTest()
	var (
		get       = "GET"
		statusURI = "localhost:9873/avaliable"
		data      interface{}
	)

	_, err := RemoteCall(get, statusURI, data)
	assert.Nil(t, err)
}

func TestResponseSuccess(t *testing.T) {
	initHttpTest()

	var (
		get  = "GET"
		uri  = "localhost:9872/terminate"
		data interface{}
	)

	ret := ResponseSuccess(get, uri, data)
	assert.Nil(t, ret)
}

func Test_parseMessage(t *testing.T) {
	var (
		responseOK = `{"suc": true, 
					"msg": "/var/keentune/bench/scripts/benchmark/wrk/wrk_http_long.py"
					}`
		failMsg      = "server is not ok"
		responseFail = `{"suc": false, 
					"msg": "server is not ok"
					}`
		responseErr = `"suc: true"`
		errMsg      = `"suc: true"`
	)

	ret := parseMessage([]byte(responseOK))
	assert.Empty(t, ret)

	ret = parseMessage([]byte(responseFail))
	assert.Equal(t, failMsg, ret)

	ret = parseMessage([]byte(responseErr))
	assert.Equal(t, errMsg, ret)
}
