/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0

* Package utils for daemon, this package contains the calculator, parsejson, utils, http. The package contains the parsing of the json file, calculate the suffix expressions, and the final result is saved at the top of the stack.Implement the device ip acquisition and validity judgment, print content string by color.
 */

package utils

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func TestRuntime(t *testing.T) {
	res := Runtime(time.Now())
	assert.Contains(t, res.Desc, "use time")
}

func TestFluctuation(t *testing.T) {
	list := []float32{float32(1.2), float32(1.3)}
	avg := float32(1.0)
	res := Fluctuation(list, avg)
	want := "fluctuation = [+20.00%, +30.00%]"
	assert.Equal(t, want, res)
}

func TestIncreaseRatio(t *testing.T) {
	last := float32(1.0)
	later := float32(1.2)
	ret := IncreaseRatio(later, last)

	assert.NotNil(t, ret)
}

func TestColorString(t *testing.T) {
	want := "\u001B[1;40;32mok\u001B[0m"
	ret := ColorString("green", "ok")
	assert.Equal(t, want, ret)
}

func TestGetExternalIP(t *testing.T) {
	_, err := GetExternalIP()
	assert.Equal(t, nil, err)
}

func TestCheckIPValidity(t *testing.T) {
	ret, _, _ := CheckIPValidity([]string{"localhost"})
	assert.Equal(t, []string{"localhost"}, ret)
}

func TestFormatInTable(t *testing.T) {
	want := "\n+----+---+\n|name|age|\n+----+---+\n|lili|10 |\n+----+---+"
	header := []string{"name", "age"}
	value := []string{"lili", "10"}

	data := [][]string{header, value}
	ret := FormatInTable(data)
	assert.Equal(t, want, ret)
}

func TestRemoveRepeated(t *testing.T) {
	want := []string{"a", "b", "c"}
	ret := RemoveRepeated(want)
	assert.Equal(t, ret, want)
}

func TestDeepCopy(t *testing.T) {
	dest := make(map[string]string)
	var src struct {
		name string `json:"name"`
	}

	src.name = "hello"
	ret := DeepCopy(dest, src)
	assert.NotNil(t, ret)
}

func TestPing(t *testing.T) {
	ret := Ping("localhost", "80")
	assert.Nil(t, ret)
}

func Test_isIP(t *testing.T) {
	var (
		ip     = "127.0.0.1"
		expect = true
	)

	got := isIP(ip)
	assert.Equal(t, expect, got)
}

func Test_decorateString(t *testing.T) {
	var (
		shiftStr = "hello"
		maxes    = []int{4}
		expect   = "hello+----+\n"
	)

	got := decorateString(maxes, shiftStr)
	assert.Equal(t, expect, got)
}

func Test_dail(t *testing.T) {
	var (
		ip   = "localhost"
		port = "9871"
	)

	err := dail(ip, port)
	assert.Nil(t, err)
}

func Test_pingTong(t *testing.T) {
	var (
		ip = "localhost"
	)

	tong := pingTong(ip)
	assert.Equal(t, true, tong)
}

func Test_isPingExist(t *testing.T) {
	var (
		expect = true
	)

	exist := isPingExist()
	assert.Equal(t, expect, exist)
}
