/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0
 */
package modules

import (
	"encoding/json"
	"fmt"
	"keentune/daemon/common/utils/http"
)

// application
const (
	myConfApp = "MySQL"
)

// avoid domain
const (
	selinuxDomain = "selinux"
)

const (
	myConfBackupFile = "/etc/my.cnf"
)

const backupAllErr = "All of the domain backup failed"

var (
	backupENVNotMetFmt = "Can't find %v, please check %v is installed"
)

var unavailableDomainReason = map[string]string{
	myConfDomain:  fmt.Sprintf(backupENVNotMetFmt, myConfBackupFile, myConfApp),
	selinuxDomain: fmt.Sprintf("unvaliable domain '%v': %v", selinuxDomain, notSupportRecommend),
}

func (tuner *Tuner) backup() error {
	err := tuner.concurrent("backup")
	if tuner.Flag == JobProfile {
		return err
	}

	if err != nil {
		return err
	}

	if tuner.backupWarning != "" {
		tuner.deleteUnAVLParams()
	}

	return nil
}

func callBackup(method, url string, request interface{}) (map[string]map[string]string, string, int) {
	var response map[string]Response
	resp, err := http.RemoteCall(method, url, request)
	if err != nil {
		return nil, err.Error(), FAILED
	}

	if err = json.Unmarshal(resp, &response); err != nil {
		return nil, err.Error(), FAILED
	}

	req, ok := request.(map[string]interface{})
	if !ok {
		return nil, "assert request type to map error", FAILED
	}

	var unAVLParam = make(map[string]map[string]string)
	var failedInfo string
	for domain, param := range req {
		domainRes := response[domain]
		parameter, assertOK := param.(map[string]interface{})
		if !assertOK {
			failedInfo += fmt.Sprintf("assert request '%v' param list failed", domain)
			continue
		}

		total := len(domainRes.ParamsResult)
		if total == 0 {
			defReason := unavailableDomainReason[domain]
			unAVLParam[domain] = map[string]string{
				domain: defReason + domainRes.Msg,
			}

			continue
		}

		var failed int
		params := domainRes.ParamsResult
		for name := range parameter {
			res, exists := params[name]
			if !exists || !res.Suc {
				failed++

				msg := res.Msg
				if msg == "" {
					msg = domainRes.Msg
				}

				_, find := unAVLParam[domain]

				if !find {
					unAVLParam[domain] = make(map[string]string)
				}

				unAVLParam[domain][name] = fmt.Sprintf("'%v' backup failed, %v", name, res.Msg)
			}
		}

		if failed == total && !domainRes.Suc {
			unAVLParam[domain] = map[string]string{
				domain: domainRes.Msg,
			}
		}

	}

	return unAVLParam, failedInfo, SUCCESS
}
