package modules

import (
	"fmt"
	"github.com/stretchr/testify/assert"
	"keentune/daemon/common/config"
	"keentune/daemon/common/log"
	"sync"
	"testing"
	"time"
)

func TestTuner_RunBenchmark(t *testing.T) {
	var (
		tuner = &Tuner{
			Benchmark: Benchmark{
				Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
				FilePath: "benchmark/wrk/wrk_http_long.py",
				SortedItems: []string{
					"Latency_90",
					"Latency_99",
					"Requests_sec",
					"Transfer_sec",
				},
				round:   1,
				LogName: log.ParamTune,
				Items: map[string]ItemDetail{
					"Requests_sec": {
						Negative: false,
						Weight:   100,
						Strict:   false,
					},
					"Transfer_sec": {
						Negative: false,
						Weight:   0,
						Strict:   false,
					},
					"Latency_90": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
					"Latency_99": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
				},
			},
		}
		num    = 1
		isBase = true
	)
	InitNodesTest()
	InitHttpChan()
	_, _, err := tuner.RunBenchmark(num, isBase)
	assert.NotNil(t, err)
}

func TestTuner_doBenchmark(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	var (
		tuner = &Tuner{
			Benchmark: Benchmark{
				Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
				FilePath: "benchmark/wrk/wrk_http_long.py",
				SortedItems: []string{
					"Latency_90",
					"Latency_99",
					"Requests_sec",
					"Transfer_sec",
				},
				round:   1,
				LogName: log.ParamTune,
				Items: map[string]ItemDetail{
					"Requests_sec": {
						Negative: false,
						Weight:   100,
						Strict:   false,
					},
					"Transfer_sec": {
						Negative: false,
						Weight:   0,
						Strict:   false,
					},
					"Latency_90": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
					"Latency_99": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
				},
			},
		}

		num = 1
	)
	tuner.doBenchmark(num)
	assert.Nil(t, nil)
}

func Test_getBenchReq(t *testing.T) {
	var (
		cmd = "python3 benchmark/wrk/wrk_http_long.py localhost"
		id  = 1
	)
	config.RealLocalIP = "localhost"
	config.KeenTune = new(config.KeentunedConf)
	config.KeenTune.Port = "9871"

	ret := getBenchReq(cmd, id)
	assert.NotNil(t, ret)

}

func Test_doOneBench(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	var (
		cmd         = "python3 benchmark/wrk/wrk_http_long.py localhost"
		wg          = sync.WaitGroup{}
		benchResult = make([]*string, len(config.KeenTune.BenchIPMap))
		sc          = NewSafeChan()

		scores = make([]map[string][]float32, len(config.KeenTune.BenchGroup))

		req = request{
			host:    fmt.Sprintf("%s:%s", "localhost", "9874"),
			id:      0,
			ipIndex: 1,
			body:    getBenchReq(cmd, 1),
			groupID: 0,
		}
		num = 1
	)

	defer sc.SafeStop()
	wg.Add(1)
	benchResult[0] = new(string)
	doOneBench(&wg, req, scores, benchResult, num, sc)
	assert.Nil(t, nil)
}

func TestBenchmark_getScore(t *testing.T) {
	var (
		bench = &Benchmark{
			Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
			FilePath: "benchmark/wrk/wrk_http_long.py",
			SortedItems: []string{
				"Latency_90",
				"Latency_99",
				"Requests_sec",
				"Transfer_sec",
			},
			round:   1,
			LogName: log.ParamTune,
			Items: map[string]ItemDetail{
				"Requests_sec": {
					Negative: false,
					Weight:   100,
					Strict:   false,
				},
				"Transfer_sec": {
					Negative: false,
					Weight:   0,
					Strict:   false,
				},
				"Latency_90": {
					Negative: true,
					Weight:   0,
					Strict:   true,
				},
				"Latency_99": {
					Negative: true,
					Weight:   0,
					Strict:   true,
				},
			},
		}
		scores = map[string][]float32{
			"Latency_90":   {float32(8590)},
			"Latency_99":   {float32(17230)},
			"Requests_sec": {float32(117274.68)},
			"Transfer_sec": {float32(468730000)},
		}

		start     = time.Now()
		benchTime = new(time.Duration)
	)

	_, _, err := bench.getScore(scores, start, benchTime)
	assert.Nil(t, err)

}

func TestBenchmark_SendScript(t *testing.T) {
	var (
		bench = &Benchmark{
			FilePath: "benchmark/wrk/wrk_http_long.py",
		}
		sendTime = new(time.Duration)
		host     = "localhost:9874"
	)

	InitNodesTest()

	_, _, err := bench.SendScript(sendTime, host)
	assert.Nil(t, err)
}

func TestTuner_analyseScore(t *testing.T) {
	var (
		tuner = &Tuner{
			baseScore: map[string]ItemDetail{
				"Requests_sec": {
					Baseline: []float32{float32(117100)},
				},
				"Transfer_sec": {
					Baseline: []float32{float32(468730100)},
				},
				"Latency_90": {
					Baseline: []float32{float32(8580)},
				},
				"Latency_99": {
					Baseline: []float32{float32(17130)},
				},
			},
			Benchmark: Benchmark{
				SortedItems: []string{
					"Latency_90",
					"Latency_99",
					"Requests_sec",
					"Transfer_sec",
				},
				round:   1,
				LogName: log.ParamTune,
				Items: map[string]ItemDetail{
					"Requests_sec": {
						Negative: false,
						Weight:   100,
						Strict:   false,
					},
					"Transfer_sec": {
						Negative: false,
						Weight:   0,
						Strict:   false,
					},
					"Latency_90": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
					"Latency_99": {
						Negative: true,
						Weight:   0,
						Strict:   true,
					},
				},
			},
		}

		scores = map[string][]float32{
			"Latency_90":   {float32(8590)},
			"Latency_99":   {float32(17230)},
			"Requests_sec": {float32(117274.68)},
			"Transfer_sec": {float32(468730000)},
		}

		start       = time.Now()
		benchResult = make([]*string, 1)
	)
	benchResult[0] = new(string)

	err := tuner.analyseScore(scores, benchResult, start)
	assert.Nil(t, err)
}

func Test_parseScore(t *testing.T) {
	var (
		sc   = NewSafeChan()
		body = `{"suc": true, "msg": "Benchmark is Runing"}`
		req  = request{
			host:    "localhost:9874",
			ipIndex: 1,
		}
	)
	InitNodesTest()
	InitHttpChan()

	defer sc.SafeStop()
	_, err := parseScore([]byte(body), req, sc)
	assert.NotNil(t, err)
}

func Test_closeChan(t *testing.T) {
	InitHttpChan()
	closeChan()
	assert.Nil(t, nil)
}
