/*
* Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
* SPDX-License-Identifier: MulanPSL-2.0
 */
package modules

import (
	"encoding/json"
	"fmt"
	"keentune/daemon/common/config"
	"keentune/daemon/common/utils/http"
	"os"
	"strings"
	"sync"
)

// StopSig ...
var StopSig chan os.Signal

// Status code
const (
	// SUCCESS status code
	SUCCESS = iota + 1
	WARNING
	FAILED
)

const multiSeparator = "*#++#*"

// backup doesn't exist
const (
	// BackupNotFound error information
	BackupNotFound = "Can not find backup file"
	FileNotExist   = "do not exists"
	NoNeedRollback = "don't need rollback"
	NoBackupFile   = "No such file"

	disableRollback = "rollback all is disable in this version"
	uninitialized   = "domain uninitialized"
)

const (
	bakUri   = "backup"
	rbUri    = "rollback"
	rbAllUri = "rollbackall"
)

var isIgnoredRollbackErr = map[string]struct{}{
	BackupNotFound:  {},
	FileNotExist:    {},
	NoNeedRollback:  {},
	NoBackupFile:    {},
	disableRollback: {},
	uninitialized:   {},
}

// Response identical target response struct
type Response struct {
	ParamsResult map[string]ParamResult `json:"param_result"`
	Suc          bool                   `json:"suc"`
	Msg          string                 `json:"msg"`
}

type ParamResult struct {
	Suc bool        `json:"suc"`
	Val interface{} `json:"value,omitempty"`
	Msg interface{} `json:"msg"`
}

func (tuner *Tuner) isInterrupted() bool {
	select {
	case <-StopSig:
		tuner.rollback()
		return true
	default:
		return false
	}
}

// Rollback ...
func Rollback(logName string, callType string) (string, error) {
	tune := new(Tuner)
	tune.logName = logName
	tune.initRB()
	var err error
	if callType == "original" {
		err = tune.original()
	} else {
		err = tune.rollback()
	}

	if err != nil {
		return tune.rollbackFailure, err
	}

	return tune.rollbackDetail, nil
}

func (gp *Group) concurrentSuccess(uri string, request interface{}) (string, int) {
	wg := sync.WaitGroup{}
	var sucCount = new(int)
	var detailInfo = new(string)
	var failedInfo = new(string)
	unAVLParams := make([]map[string]map[string]string, len(gp.IPs))

	// replace rollback all to rollback
	if uri == rbAllUri {
		uri = rbUri
	}

	for index, ip := range gp.IPs {
		wg.Add(1)
		id := config.KeenTune.IPMap[ip]
		config.IsInnerApplyRequests[id] = false
		go func(index, groupID int, ip string, wg *sync.WaitGroup) {
			defer wg.Done()
			url := fmt.Sprintf("%v:%v/%v", ip, gp.Port, uri)
			var msg string
			var status int
			if uri != "backup" {
				msg, status = callRollback("POST", url, request)
			} else {
				unAVLParams[index-1], msg, status = callBackup("POST", url, request)
			}

			switch status {
			case SUCCESS, WARNING:
				*sucCount++
				if msg != "" {
					*detailInfo += fmt.Sprintf("\n\tGroup %v Node %v: %v %v", groupID, index, ip, msg)
					break
				}

				*detailInfo += fmt.Sprintf("Group %v Node %v: %v ", groupID, index, ip)
			case FAILED:
				*failedInfo += fmt.Sprintf("\tGroup %v Node %v: %v\n%v\n", groupID, index, ip, msg)
			}

		}(index+1, gp.GroupNo, ip, &wg)
	}

	wg.Wait()

	if uri == bakUri {
		if *failedInfo != "" {
			return *failedInfo, FAILED
		}

		warningInfo, status := gp.deleteUnAVLConf(unAVLParams)
		gp.updateRBDomains()
		return warningInfo, status
	}

	if *sucCount == len(gp.IPs) {
		return *detailInfo, SUCCESS
	}

	return *failedInfo, FAILED
}

func callRollback(method string, url string, request interface{}) (string, int) {
	resp, err := http.RemoteCall(method, url, request)
	if err != nil {
		if strings.Contains(err.Error(), "connection refused") {
			return "server is offline", FAILED
		}

		return err.Error(), FAILED
	}

	var response map[string]Response

	if err = json.Unmarshal(resp, &response); err != nil {
		return string(resp), FAILED
	}

	var parseFailedMsg string
	var warnCount int

	for domain, res := range response {
		if parseStatusCode(res.Msg) == WARNING {
			warnCount++
			continue
		}

		if !res.Suc {
			parseFailedMsg += fmt.Sprintf("\n\t\t[%v] domain rollback failed: %v", domain, res.Msg)
			for name, v := range res.ParamsResult {
				if !v.Suc && v.Msg != "" {
					parseFailedMsg += fmt.Sprintf("\n\t\t\t'%v' rollback failed: %v", name, v.Msg)
				}
			}
			continue
		}

		if res.Msg != "" {
			parseFailedMsg += fmt.Sprintf("\n\t\t[%v] domain rollback successfully, %v", domain, res.Msg)
			continue
		}

		parseFailedMsg += fmt.Sprintf("\n\t\t[%v] domain rollback successfully", domain)
	}

	if warnCount == len(response) {
		return "No domain needs to be rolled back", WARNING
	}

	return parseFailedMsg, SUCCESS
}

func parseStatusCode(msg interface{}) int {
	switch info := msg.(type) {
	case map[string]interface{}:
		var count int
		for _, value := range info {
			message := fmt.Sprint(value)
			if assertIgnoreRollbackErr(message) {
				count++
			}
		}

		if count == len(info) && count > 0 {
			return WARNING
		}
		return SUCCESS
	case string:
		if assertIgnoreRollbackErr(info) {
			return WARNING
		}
		return SUCCESS
	case interface{}:
		if info == nil {
			return SUCCESS
		}
		return WARNING
	}

	return SUCCESS
}

func parseMsg(originMsg interface{}) string {
	var resp map[string]struct {
		Suc bool   `json:"suc"`
		Msg string `json:"msg"`
	}

	msg, _ := json.Marshal(originMsg)
	err := json.Unmarshal(msg, &resp)
	if err != nil {
		return string(msg)
	}

	var retMsg string

	for domain, info := range resp {
		if info.Suc {
			continue
		}

		replaced := strings.ReplaceAll(info.Msg, "\n", "\n\t\t\t")

		retMsg += fmt.Sprintf("\t\t[%s]\t%v\n", domain, replaced)
	}

	if len(retMsg) != 0 {
		return retMsg
	}

	return string(msg)
}

func (tuner *Tuner) original() error {
	tuner.rollbackReq = map[string]interface{}{
		"domains": []string{},
		"all":     true,
	}
	return tuner.concurrent(rbAllUri)
}

func assertIgnoreRollbackErr(msg string) bool {
	for key := range isIgnoredRollbackErr {
		if strings.Contains(msg, key) {
			return true
		}
	}

	return false
}
