package modules

import (
	"github.com/stretchr/testify/assert"
	"keentune/daemon/common/config"
	"testing"
)

func TestTuner_init(t *testing.T) {
	var (
		tuner = &Tuner{
			Name:         "aaa",
			Algorithm:    "hord",
			MAXIteration: 1,
		}
	)

	InitNodesTest()
	InitHttpChan()

	err := tuner.init()
	assert.NotNil(t, err)
}

func TestTuner_dump(t *testing.T) {
	InitNodesTest()

	var (
		tuner = &Tuner{
			Name:         "aaa",
			Algorithm:    "hord",
			MAXIteration: 1,
			Group: []Group{
				{
					Dump: Configuration{
						Round: 1,
					},
				},
			},
		}
		option = processOpt
	)

	config.KeenTune.ExecDump = true

	tuner.dump(option)
	assert.Nil(t, nil)
}

func TestTuner_baseline(t *testing.T) {
	InitNodesTest()
	InitHttpChan()

	var (
		tuner = &Tuner{
			Name:         "aaa",
			Algorithm:    "hord",
			MAXIteration: 1,
			Group: []Group{
				{
					Dump: Configuration{
						Round: 1,
					},
				},
			},
			Benchmark: Benchmark{
				Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
				FilePath: "benchmark/wrk/wrk_http_long.py",
			},
		}
	)

	config.KeenTune.BaseDump = true

	ret := tuner.baseline()
	assert.NotNil(t, ret)

}

func TestTuner_brainInit(t *testing.T) {
	InitNodesTest()

	var (
		baseScore = map[string]ItemDetail{
			"Requests_sec": {
				Baseline: []float32{float32(117100)},
				Negative: false,
				Weight:   100,
				Strict:   false,
			},
			"Transfer_sec": {
				Baseline: []float32{float32(468730100)},
				Negative: false,
				Weight:   0,
				Strict:   false,
			},
			"Latency_90": {
				Baseline: []float32{float32(8580)},
				Negative: true,
				Weight:   0,
				Strict:   true,
			},
			"Latency_99": {
				Baseline: []float32{float32(17130)},
				Negative: true,
				Weight:   0,
				Strict:   true,
			},
		}
		tuner = &Tuner{
			Name:         "aaa",
			Algorithm:    "hord",
			MAXIteration: 10,
			Flag:         "tuning",
			Group: []Group{
				{
					Dump: Configuration{
						Round: 1,
					},
					Params: []config.DBLMap{
						{},
						{
							"sysctl": {
								"vm.dirty_ratio": map[string]interface{}{
									"domain": "sysctl",
									"range": []int{
										0, 100,
									},
									"dtype": "int",
									"step":  1,
									"base":  81,
									"value": 5,
								},
							},
						},
					},
				},
			},

			Benchmark: Benchmark{
				Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
				FilePath: "benchmark/wrk/wrk_http_long.py",
			},

			baseScore: baseScore,

			BrainParam: []Parameter{
				{
					DomainName: "sysctl",
					ParaName:   "vm.dirty_ratio@group-1",
					Scope: []interface{}{
						0,
						100,
					},
					Dtype: "int",
					Step:  1,
				},
			},
		}
	)

	defer func() { tuner.end() }()

	config.KeenTune.BaseDump = true

	ret := tuner.brainInit()
	assert.Nil(t, ret)
}

func Test_requestInit(t *testing.T) {
	var (
		baseScore = map[string]ItemDetail{
			"Requests_sec": {
				Baseline: []float32{float32(117100)},
				Negative: false,
				Weight:   100,
				Strict:   false,
			},
			"Transfer_sec": {
				Baseline: []float32{float32(468730100)},
				Negative: false,
				Weight:   0,
				Strict:   false,
			},
			"Latency_90": {
				Baseline: []float32{float32(8580)},
				Negative: true,
				Weight:   0,
				Strict:   true,
			},
			"Latency_99": {
				Baseline: []float32{float32(17130)},
				Negative: true,
				Weight:   0,
				Strict:   true,
			},
		}
		param = []Parameter{
			{
				DomainName: "sysctl",
				ParaName:   "vm.dirty_ratio@group-1",
				Scope: []interface{}{
					0,
					100,
				},
				Dtype: "int",
				Value: 80,
				Step:  1,
				Base:  6,
			},
		}
		requireConf = map[string]interface{}{
			"algorithm":      "",
			"iteration":      2,
			"name":           "aaa",
			"parameters":     param,
			"baseline_score": baseScore,
		}
		job = "aaa"
	)

	ret := requestInit(requireConf, job)
	assert.NotNil(t, ret)
}

func TestTuner_rollback(t *testing.T) {
	InitNodesTest()

	var (
		tuner = &Tuner{
			Group: []Group{
				{
					IPs:             []string{"localhost"},
					Port:            "9873",
					rollbackDomains: []string{"sysctl"},
				},
			},
		}
	)

	ret := tuner.rollback()
	assert.Nil(t, ret)
}

func TestTuner_concurrent(t *testing.T) {
	InitNodesTest()

	var (
		tuner = &Tuner{
			Group: []Group{
				{
					IPs:             []string{"localhost"},
					Port:            "9873",
					rollbackDomains: []string{"sysctl"},
				},
			},
		}

		uri = rbUri
	)

	ret := tuner.concurrent(uri)
	assert.Nil(t, ret)
}

func TestTuner_saveBrainInit(t *testing.T) {
	InitNodesTest()

	var (
		tuner = &Tuner{
			Name: "aaa",
			Flag: "tuning",
			Group: []Group{
				{
					IPs:             []string{"localhost"},
					Port:            "9873",
					rollbackDomains: []string{"sysctl"},
				},
			},
			BrainParam: []Parameter{
				{
					DomainName: "sysctl",
					ParaName:   "vm.dirty_ratio@group-1",
					Scope: []interface{}{
						0,
						100,
					},
					Dtype: "int",
					Value: 80,
					Step:  1,
				},
			},
			baseScore: map[string]ItemDetail{
				"Requests_sec": {
					Baseline: []float32{float32(117100)},
				},
				"Transfer_sec": {
					Baseline: []float32{float32(468730100)},
				},
				"Latency_90": {
					Baseline: []float32{float32(8580)},
				},
				"Latency_99": {
					Baseline: []float32{float32(17130)},
				},
			},
		}
	)

	tuner.saveBrainInit()
	assert.Nil(t, nil)
}

