package modules

import (
	"github.com/stretchr/testify/assert"
	"keentune/daemon/common/config"
	"testing"
)

func Test_updateParameter(t *testing.T) {
	var (
		partial = &Parameter{}

		total = &Parameter{
			Dtype: "int",
			Scope: []interface{}{
				0,
				88,
			},
			Step: 2,
		}
	)

	updateParameter(partial, total)
	assert.Nil(t, nil)
}

func TestUpdateParams(t *testing.T) {
	InitFileTest()
	var userParam = config.DBLMap{
		"sysctl": {
			"vm.page-cluster": map[string]interface{}{
				"domain": "sysctl",
				"range": []int{
					0, 8,
				},
				"dtype": "int",
				"step":  1,
				"base":  "6",
				"value": 5,
			},
		},
	}
	UpdateParams(userParam)
	assert.Nil(t, nil)
}

func Test_modifyParam(t *testing.T) {
	var (
		originMap = &map[string]interface{}{
			"domain": "sysctl",
			"range": []int{
				0, 8,
			},
			"step":  0,
			"base":  "6",
			"value": 5,
		}
		compareMap = map[string]interface{}{
			"vm.page-cluster": map[string]interface{}{
				"domain": "sysctl",
				"range": []int{
					0, 8,
				},
				"dtype": "int",
				"step":  1,
				"base":  "6",
				"value": 5,
			},
		}
		paramName = "vm.page-cluster"
	)

	err := modifyParam(originMap, compareMap, paramName)
	assert.Nil(t, err)
}

func Test_getExtremeValue(t *testing.T) {
	var (
		macros             = []string{"#!cpu_core#"}
		detectedMacroValue = map[string]string{}

		macroString = "MAX(#!cpu_core#*2, 8)"
		ip          = "localhost"
	)

	InitNodesTest()
	value, err := getExtremeValue(macros, detectedMacroValue, macroString, ip)
	assert.NotEmpty(t, value)
	assert.Nil(t, err)
}

func Test_convertString(t *testing.T) {
	var (
		macroString = "MAX(#!cpu_core#*2, 16)"
		macroMap    = map[string]string{
			"#!cpu_core#": "4",
		}
	)
	got, symbol, got2 := convertString(macroString, macroMap)
	assert.NotNil(t, got2)
	assert.NotEmpty(t, got)
	assert.Equal(t, "MAX", symbol)
}

func Test_getMacroValue(t *testing.T) {
	var (
		macros             = []string{"#!cpu_core#"}
		detectedMacroValue = map[string]string{}

		ip = "localhost"
	)

	err := getMacroValue(macros, detectedMacroValue, ip)
	assert.Nil(t, err)
}

func TestConvertConfFileToJson(t *testing.T) {
	var (
		fileName = "/etc/keentune/keentuned/profile/application/mysql.conf"
		ip       = []string{
			"localhost",
		}
	)
	_, _, err := ConvertConfFileToJson(fileName, ip...)
	assert.Nil(t, err)
}

func Test_readConfFile(t *testing.T) {
	var (
		fileName = "/etc/keentune/keentuned/profile/application/mysql.conf"
	)
	got, _ := readConfFile(fileName)
	assert.NotEmpty(t, got)
}

func Test_changeMapSliceToDBLMap(t *testing.T) {
	var (
		domainMap = map[string][]map[string]interface{}{
			"sysctl": []map[string]interface{}{
				{
					"vm.page-cluster": map[string]interface{}{
						"domain": "sysctl",
						"range": []int{
							0, 8,
						},
						"dtype": "int",
						"step":  1,
						"base":  "6",
						"value": 5,
					},
				},
			},
		}
		abnormal ABNLResult
	)
	_, got1, err := changeMapSliceToDBLMap(domainMap, abnormal)
	assert.NotNil(t, got1)
	assert.Nil(t, err)
}

func Test_parseConfStrToMapSlice(t *testing.T) {
	var (
		replacedStr = `[main]
include = ecs-guest.conf

[net]
XPS = different
# XPS i.e. different, off
RPS = different
# RPS i.e. different, off
smp_affinity = dedicated
# smp_affinity i.e. different, dedicated, off

[kernel_mem]
code_hugepage = 1
# code_hugepage i.e. 0, 1, 2, 3
TLBI = "recommend: TLB range improvement. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/TLBI%20-%20TLB%20range%E4%BC%98%E5%8C%96.md"

[kernel_io]
atomic_write = "recommend：Close MySQL double write and use 16K atomic write can improve write request performance. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/16K%E5%8E%9F%E5%AD%90%E5%86%99.md"
ext4_fast_commit = "recommend: fast commit for ext4. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/EXT4%20Fast%20Commit.md"

[kernel_sec]
E0PD = "recommend: Please use E0PD security feature to avoid meltdown attack. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/E0PD.md"

[mysql_compile]
LSE = "recommend: LSE appplies GCC compile promotion. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/LSE%E6%8C%87%E4%BB%A4%E9%9B%86%E7%BC%96%E8%AF%91%E4%BC%98%E5%8C%96.md"
LTO = "recommend: Link Time Optimization. Please refer to https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/LTO.md"
PGO = "recommend：PGO can produce optimal code by using application runtime data. Please refer to: https://gitee.com/anolis/community/blob/master/sig/ARM_SIG/content/profile_features/PGO.md"

[sysctl]
kernel.sched_migration_cost_ns = 5000
kernel.sched_min_granularity_ns = 3000000
kernel.sched_wakeup_granularity_ns = 4000000`
		fileName = "mysql.conf"
		abnormal = &ABNLResult{}
		ip       = []string{"localhost"}
	)
	_, got1, got2 := parseConfStrToMapSlice(replacedStr, fileName, abnormal, ip...)
	assert.NotNil(t, got1)
	assert.NotNil(t, got2)
}

func Test_collectConfVariables(t *testing.T) {
	var (
		pureLine    = "isolated_cores_expanded=${f:cpulist_unpack:${isolated_cores}}"
		variableMap = map[string]string{
			"isolated_cores": "5",
		}
		variableReq = map[string]interface{}{}
	)
	collectConfVariables(pureLine, variableMap, variableReq)
	assert.Nil(t, nil)
}

func Test_parseIncludeConf(t *testing.T) {
	var (
		pureLine = "include:network-latency"
		abnormal = &ABNLResult{}
		ip       = []string{"localhost"}
	)
	InitFileTest()
	got := parseIncludeConf(pureLine, abnormal, ip...)
	assert.NotNil(t, got)
}

func Test_mergedMapSlice(t *testing.T) {
	var (
		domainMap = map[string][]map[string]interface{}{
			"sysctl": {
				{
					"vm.page-cluster": map[string]interface{}{
						"domain": "sysctl",
						"range": []int{
							0, 8,
						},
						"dtype": "int",
						"step":  1,
						"base":  "6",
						"value": 5,
					},
				},
			},
		}
		includeMap = map[string][]map[string]interface{}{
			"cri_sysctl": {
				{
					"vm.page-cluster": map[string]interface{}{
						"domain": "sysctl",
						"range": []int{
							0, 8,
						},
						"dtype": "int",
						"step":  1,
						"base":  "6",
						"value": 5,
					},
				},
			},
		}
	)
	got := mergedMapSlice(domainMap, includeMap)
	want := len(domainMap) + len(includeMap)
	assert.Equal(t, want, len(got))
}

func Test_readLine(t *testing.T) {
	var (
		line = "innodb_buffer_pool_size : \"#!MEM_TOTAL#*1073741824 - 2*1073741824*#!CPU_CORE#\" "
		ip   = []string{
			"localhost",
		}
	)
	InitNodesTest()
	_, _, _, err := readLine(line, ip...)
	assert.Nil(t, err)
}

func Test_getParam(t *testing.T) {
	var (
		paramSlice = []string{
			"__ENV-LIMITATION",
			"#!CPU_CORE#=8 & #!MEM_TOTAL# ∈ [29, 32]",
		}
		ip = []string{
			"localhost",
		}
	)
	InitNodesTest()
	_, _, _, err := getParam(paramSlice, ip...)
	assert.Nil(t, err)
}

func Test_genParam(t *testing.T) {
	var (
		valueStr  = "37355520  # this is bit unit"
		paramName = "kernel.threads-max"
	)
	got := genParam(valueStr, paramName)
	assert.NotNil(t, got)
}

func Test_isRecommend(t *testing.T) {
	var (
		valueStr  = "recommend: Please use E0PD instead of BHB. Please refer to https://gitee.com/anolis/keentuned/blob/dev-1.3.3/docs/profile_features/E0PD.md"
		paramName = "E0PD"
	)
	got, got1 := isRecommend(valueStr, paramName)
	assert.NotEmpty(t, got)
	assert.Equal(t, true, got1)

}

func Test_replaceEqualSign(t *testing.T) {
	var (
		origin = "code_hugepage = 1"
	)
	got := replaceEqualSign(origin)
	assert.NotEmpty(t, got)
}

func TestConvertToSequentialDict(t *testing.T) {
	var (
		fileName = "/etc/keentune/keentuned/profile/default.conf"
	)
	got, err := ConvertToSequentialDict(fileName)
	assert.NotEmpty(t, got)
	assert.Nil(t, err)
}

