package modules

import (
	"encoding/json"
	"fmt"
	"keentune/daemon/common/log"
	"testing"

	"github.com/stretchr/testify/assert"
)

func TestTuner_Set(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Setter: Setter{
			Group: []bool{true},
			ConfFile: []string{"/etc/keentune/keentuned/profile/basic" +
				"/cpu_high_load.conf"},
		},
		Flag: JobProfile,
	}

	tuner.Set()
	assert.Nil(t, nil)
}

func TestTuner_showReco(t *testing.T) {
	InitNodesTest()
	tuner := &Tuner{
		Setter: Setter{
			Group:       []bool{true},
			ConfFile:    []string{"cpu_high_load.conf"},
			recommend:   "this is recommend Test",
			initWarning: "This is initWarning Test",
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
	}

	tuner.showReco()
	assert.Nil(t, nil)
}

func TestTuner_showPrefixReco(t *testing.T) {
	InitNodesTest()
	tuner := &Tuner{
		Setter: Setter{
			prefixReco: "This is prefixReco Test",
		},
		logName: log.ProfSet,
	}

	tuner.showPrefixReco()
	assert.Nil(t, nil)
}

func TestTuner_updateActive(t *testing.T) {
	defer UpdateActiveFile("/var/keentune/keentuned/profile/active.conf", []byte{})

	InitNodesTest()
	tuner := &Tuner{
		Setter: Setter{
			Group:    []bool{true},
			ConfFile: []string{"cpu_high_load.conf"},
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
	}

	err := tuner.updateActive()
	assert.Nil(t, err)
}

func TestTuner_prepareBeforeSet(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	var (
		paramStr = `{
	"sysctl": {
		"vm.page-cluster": {
			"dtype": "int"
		},
		"vm.stat_interval": {
			"dtype": "int"
		},
		"vm.swappiness": {
			"dtype": "int"
		},
		"vm.vfs_cache_pressure": {
			"dtype": "int"
		},
		"vm.watermark_scale_factor": {
			"dtype": "int"
		},
		"vm.zone_reclaim_mode": {
			"dtype": "string"
		}
	}
}`
		params map[string]interface{}
	)

	json.Unmarshal([]byte(paramStr), &params)

	tuner := &Tuner{
		Setter: Setter{
			Group:    []bool{true},
			ConfFile: []string{"cpu_high_load.conf"},
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
		Group: []Group{
			{
				rollbackDomains: []string{"sysctl"},
				MergedParam:     params,
			},
		},
	}

	err := tuner.prepareBeforeSet()
	assert.Nil(t, err)
}

func TestUpdateActiveFile(t *testing.T) {
	var (
		fileName = "/var/keentune/keentuned/profile/active.conf"
		info     []byte
	)

	ret := UpdateActiveFile(fileName, info)
	assert.Nil(t, ret)
}

func TestTuner_SetDefault(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Setter: Setter{
			Group: []bool{true},
			ConfFile: []string{"/etc/keentune/keentuned/profile/basic" +
				"/cpu_high_load.conf"},
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
		Group: []Group{
			{
				IPs: []string{
					"localhost",
				},
				Port: "9873",
				initDomains: []string{
					"sysctl",
				},
				GroupNo: 1,
			},
		},
	}

	_, _, err := tuner.SetDefault()
	assert.Nil(t, err)

}

func TestTuner_initDefaultDomain(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Setter: Setter{
			Group:    []bool{true},
			ConfFile: []string{"cpu_high_load.conf"},
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
		Group: []Group{
			{
				IPs: []string{
					"localhost",
				},
				Port: "9873",
				initDomains: []string{
					"sysctl",
				},
				GroupNo: 1,
			},
		},
	}

	_, err := tuner.initDefaultDomain()
	assert.NotNil(t, err)
}

func TestTuner_updateGroup(t *testing.T) {
	InitNodesTest()
	tuner := &Tuner{
		Setter: Setter{
			Group:    []bool{true},
			ConfFile: []string{"cpu_high_load.conf"},
		},
		Flag:    JobProfile,
		logName: log.ProfSet,
		Group: []Group{
			{
				IPs: []string{
					"localhost",
				},
				Port: "9873",
				initDomains: []string{
					"sysctl",
				},
				GroupNo: 1,
			},
		},
	}

	var (
		paramStr = `{
	"sysctl": {
		"vm.page-cluster": {
			"dtype": "int"
		},
		"vm.stat_interval": {
			"dtype": "int"
		},
		"vm.swappiness": {
			"dtype": "int"
		},
		"vm.vfs_cache_pressure": {
			"dtype": "int"
		},
		"vm.watermark_scale_factor": {
			"dtype": "int"
		},
		"vm.zone_reclaim_mode": {
			"dtype": "string"
		}
	}
}`
		params map[string]map[string]interface{}
	)

	json.Unmarshal([]byte(paramStr), &params)

	tuner.updateGroup(params)
	assert.Nil(t, nil)
}

func Test_handleTimeoutChan(t *testing.T) {
	err := fmt.Errorf("this is Timeout test")
	handleTimeoutChan(err)
	assert.Nil(t, nil)
}
