package modules

import (
	"encoding/json"
	"fmt"
	"keentune/daemon/common/config"
	"keentune/daemon/common/file"
	"keentune/daemon/common/log"
	"os"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
)

func TestTuner_Tune(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	defer clearTuneJob()
	item := `{
        "Requests_sec": {
          "negative": false,
          "weight": 100,
          "strict": false
        },
        "Transfer_sec": {
          "negative": false,
          "weight": 0,
          "strict": false
        },
        "Latency_90": {
          "negative": true,
          "weight": 0,
          "strict": true
        },
        "Latency_99": {
          "negative": true,
          "weight": 0,
          "strict": true
        }
      }`

	var itemDetail map[string]ItemDetail
	json.Unmarshal([]byte(item), &itemDetail)
	tuner := &Tuner{
		Name:         "aaa",
		Algorithm:    "hord",
		MAXIteration: 1,
		Iteration:    0,
		StartTime:    time.Now(),
		Benchmark: Benchmark{
			Cmd:      "python3 benchmark/wrk/wrk_http_long.py localhost",
			FilePath: "benchmark/wrk/wrk_http_long.py",
			SortedItems: []string{
				"Latency_90",
				"Latency_99",
				"Requests_sec",
				"Transfer_sec",
			},
			round:   1,
			LogName: log.ParamTune,
			Items:   itemDetail,
		},
		Step:    1,
		Flag:    "tuning",
		logName: log.ParamTune,
	}

	tuner.Tune()
	assert.Nil(t, nil)
}

func TestTuner_parseTuningError(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Name:      "ccc",
		StartTime: time.Now(),
		Flag:      "tuning",
		logName:   log.ParamTune,
	}

	err := fmt.Errorf("this is err test")
	tuner.parseTuningError(err)
}

func TestTuner_acquire(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Name:      "ccc",
		Algorithm: "hord",
		Step:      6,
		logName:   log.ParamTune,
	}
	_, err := tuner.acquire()
	assert.NotNil(t, err)
}

func TestTuner_feedback(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	var feed map[string][]float32
	var feedStr = `{"bench_score": {
		"Latency_90": [
			8590
		],
		"Latency_99": [
			17230
		],
		"Requests_sec": [
			117274.68
		],
		"Transfer_sec": [
			468730000
		]
	}}`

	json.Unmarshal([]byte(feedStr), &feed)
	tuner := &Tuner{
		Name:          "ccc",
		Algorithm:     "hord",
		Iteration:     6,
		Step:          6,
		logName:       log.ParamTune,
		feedbackScore: feed,
	}

	err := tuner.feedback()
	assert.NotNil(t, err)
}

func TestTuner_end(t *testing.T) {
	InitNodesTest()
	InitHttpChan()
	tuner := &Tuner{
		Name:    "ccc",
		logName: log.ParamTune,
		Flag:    "tuning",
	}
	tuner.end()
	assert.Nil(t, nil)
}

func Test_endTime(t *testing.T) {
	ret := endTime(int64(1234))
	assert.NotEmpty(t, ret)
}

func TestTuner_setTimeSpentDetail(t *testing.T) {
	tuner := &Tuner{
		MAXIteration: 13,
		timeSpend: timeSpend{
			init:      1,
			acquire:   1,
			apply:     1,
			send:      1,
			benchmark: 5,
			feedback:  1,
			end:       1,
			best:      3,
		},
	}
	tuner.setTimeSpentDetail(float64(3))
	assert.Nil(t, nil)
}

func TestTuner_IncreaseStep(t *testing.T) {
	tuner := &Tuner{
		Step: 1,
	}
	ret := tuner.IncreaseStep()
	assert.NotEmpty(t, ret)
}

func TestTuner_deleteUnAVLParams(t *testing.T) {
	InitNodesTest()
	tuner := &Tuner{
		Name:    "ccc",
		logName: log.ParamTune,
		Flag:    "tuning",
		Group: []Group{
			{
				UnAVLParams: map[string]map[string]string{},
			},
		},
		BrainParam: []Parameter{
			{
				DomainName: "sysctl",
				ParaName:   "vm.dirty_ratio@group-1",
				Scope: []interface{}{
					0,
					100,
				},
				Dtype: "int",
				Value: 80,
				Step:  1,
				Base:  77,
			},
		},
	}
	tuner.deleteUnAVLParams()
	assert.Nil(t, nil)
}

func clearTuneJob() {
	os.RemoveAll("/var/keentune/keentuned/tuning_workspace/aaa")
	file.DeleteRow(config.GetDumpPath(config.TuneCsv), "aaa", TuneNameIdx)
	os.Remove("/var/log/keentune/aaa.log")
}
